<?php
namespace Sale\Handlers\PaySystem;
use Bitrix\Sale;
use Bitrix\Catalog;
use Bitrix\Catalog\Mysql\CCatalogVat;
use Bitrix\Main\Application;
use Bitrix\Main\Web\Uri;
use Bitrix\Main\Error;
use Bitrix\Main\Request;
use Bitrix\Main\Type\DateTime;
use Bitrix\Main\Localization\Loc;
use Bitrix\Sale\PaySystem;
use Bitrix\Sale\Payment;
use Bitrix\Sale\PriceMaths;
class PayKeeperHandler extends PaySystem\ServiceHandler
{
    private $fiscal_cart = array();
    private $order_total = 0; //order total sum
    private $payment_form_url = "";
    private $use_delivery = FALSE;
    private $single_item_index = -1;
    private $more_then_one_item_index = -1;
	/**
	 * @param Payment $payment
	 * @param Request|null $request
	 * @return PaySystem\ServiceResult
	 */
	public function initiatePay(Payment $payment, Request $request = null)
	{
		$busValues = $this->getParamsBusValue($payment);
        $form_url = $busValues["PAYKEEPER_FORM_URL"];
        $this->setPaymentFormUrl($form_url);
        $secret = $busValues["PAYKEEPER_SECRET"];
        $orderid = ($this->getPaymentFormType() == "create") ? $payment->getOrderId() : $payment->getOrderId() . "|paykeeper_payment_system";
        $order = Sale\Order::load($orderid);
        $propertyCollection = $order->getPropertyCollection();
        $this->setOrderTotal(floatval($order->getPrice())); 
        $clientid = is_null($propertyCollection->getPayerName()) ? "" : $propertyCollection->getPayerName()->getValue();
        $client_email = is_null($propertyCollection->getUserEmail()) ? "" : $propertyCollection->getUserEmail()->getValue();
        $client_phone = is_null($propertyCollection->getPhone()) ? "" : $propertyCollection->getPhone()->getValue();
        $service_name = "paykeeper_payment_system";
        $basket = $order->getBasket();
        $item_index = 0;
        foreach ($basket as $item) {
            $taxes = array("tax" => "none", "tax_sum" => 0);
            $name = $item->getField("NAME");
            $quantity = $item->getField("QUANTITY");
            //for correctPrecision()
            if ($quantity === 1 && $this->single_item_index < 0)
                $this->single_item_index = $item_index;
            if ($quantity > 1 && $this->more_then_one_item_index < 0)
                $this->more_then_one_item_index = $item_index;
            $price = floatval($item->getField("PRICE"));
            $sum = $price*$quantity;
            $vat_rate = (int) ($item->getField("VAT_RATE")*100);
            $taxes = $this->setTaxes($sum, (float) $vat_rate);
            $this->updateFiscalCart($name, $price, $quantity, $sum, $taxes["tax"]);
            $item_index++;
        }
        $dbSaleDelivery = \Bitrix\Sale\Delivery\Services\Manager::getById($order->getField("DELIVERY_ID"));
        $delivery_name = $dbSaleDelivery["NAME"];
        $delivery_price = floatval($order->getDeliveryPrice());
        //$delivery_tax_obj = CCatalogVat::GetById($dbSaleDelivery["VAT_ID"]);
        //$delivery_tax_rate = (int)$delivery_tax_obj->Fetch()["RATE"];
        $delivery_tax_rate = 18;
        $delivery_taxes = ((int) $dbSaleDelivery["VAT_ID"] == 0) ?
                            array("tax" => "none", "tax_sum" => 0) :
                            $this->setTaxes($delivery_price, $delivery_tax_rate);
        if ($delivery_price > 0) {
            $this->setUseDelivery();
            $this->updateFiscalCart($delivery_name, $delivery_price, 1, $delivery_price, $delivery_taxes["tax"]);
        }
        $this->correctPrecision(); //correct possible difference between order sum and fiscal cart sum
        //Encode fiscal cart to utf-8 for json_encode
        $cart_encoded = array();
        foreach ($this->fiscal_cart as $product) {
            $product_ar = array();
            foreach ($product as $key => $value) {
                $enc = mb_detect_encoding($value, 'ASCII, UTF-8, windows-1251', false);
                $product_ar[$key] = ($enc == "UTF-8") ? $value : iconv($enc, "UTF-8", $value);
            }
            $cart_encoded[] = $product_ar;
        }
        $this->fiscal_cart = json_encode($cart_encoded);
		$params = array(
            "form_url" => $form_url,
            "sum" => $this->getOrderTotal(),
            "orderid" => $orderid,
            "clientid" => $clientid,
            "client_email" => $client_email,
            "client_phone" => $client_phone,
            "service_name" => $service_name,
            "phone" => $phone,
            "cart" => $this->fiscal_cart
        );
        $request = Application::getInstance()->getContext()->getRequest();
        $uriString = $request->getRequestUri();
        $uri = new Uri($uriString);
        $params["noauth_payment_active"] = strpos($uri->GetLocator(), "access=");
        $params["detail_page_active"] = strpos($uri->GetLocator(), "/personal/orders/$orderid");
        //payment form language
        switch (LANGUAGE_ID) {
            case "en":
                $pay_button = "Pay online";
                if ($params["detail_page_active"] !== false) {
                    $message = "Press pay button for redirect to bank payment gateway now.";
                    $params["detail_page_message"] = ($params["noauth_payment_active"] === false) ?
                        "Press pay button for redirect to bank payment gateway now."
                        : "Press pay button on personal account page for redirect to bank payment gateway now.";
                }
                else
                    $message = "You will be redirected to bank payment gateway now.";
                $pf_lang = "en";
                break;
            default:
                $pay_button = "Оплатить";
                if ($params["detail_page_active"] !== false) {
                    $message = "Нажмите на кнопку Оплатить для перенаправления на страницу банка.";
                    $params["detail_page_message"] = ($params["noauth_payment_active"] === false) ?
                        "Нажмите на кнопку Оплатить для перенаправления на страницу банка."
                        : "Нажмите на кнопку Оплатить в персональном разделе сайта для перенаправления на страницу банка.";
                }
                else
                    $message = "Сейчас Вы будете перенаправлены на страницу банка.";
                $pf_lang = "ru";
                break;
        }
        if (LANG_CHARSET != "utf-8" and $pf_lang == "ru")
            $message = iconv("UTF-8", LANG_CHARSET, $message);
        $params["message"] = $message;
        if ($this->getPaymentFormType() == "create") { //create form
            $params["form_type"] = "create";
            $to_hash = number_format($this->getOrderTotal(), 2, ".", "").
                       $clientid     .
                       $orderid      .
                       $service_name .
                       $client_email .
                       $client_phone .
                       $secret;
            $sign = hash ('sha256' , $to_hash);
            $params["sign"] = $sign;
            $params["lang"] = $pf_lang;
            $pay_button = (LANG_CHARSET == "UTF-8") ? $pay_button :
                      iconv("UTF-8", LANG_CHARSET, $pay_button);
            $params["pay_button"] = $pay_button;
            $js_submit_string = 'document.forms["pay_form"].submit();';
            $params["js_submit_string"] = $js_submit_string;
        }
        else { //order form
            $payment_parameters = array(
                "clientid"=>$clientid,
                "orderid"=>$orderid,
                "sum"=>$this->getOrderTotal(),
                "phone"=>$client_phone,
                "client_email"=>$client_email,
                "cart"=>$this->fiscal_cart);
            $query = http_build_query($payment_parameters);
            if( function_exists( "curl_init" )) { //using curl
                $CR = curl_init();
                curl_setopt($CR, CURLOPT_URL, $form_url);
                curl_setopt($CR, CURLOPT_POST, 1);
                curl_setopt($CR, CURLOPT_FAILONERROR, true); 
                curl_setopt($CR, CURLOPT_POSTFIELDS, $query);
                curl_setopt($CR, CURLOPT_RETURNTRANSFER, 1);
                curl_setopt($CR, CURLOPT_SSL_VERIFYPEER, 0);
                $result = curl_exec( $CR );
                $error = curl_error( $CR );
                if( !empty( $error )) {
                    $form = "<br/><span class=message>"."INTERNAL ERROR:".$error."</span>";
                    return false;
                }
                else {
                    $form = $result;
                }
                curl_close($CR);
            }
            else { //using file_get_contents
                if (!ini_get('allow_url_fopen')) {
                    $form = "<br/><span class=message>"."INTERNAL ERROR: Option allow_url_fopen is not set in php.ini"."</span>";
                }
                else
                    $form = file_get_contents($form_url, false, $context);
            }
            if ($form  == "") {
                $form = '<h3>Произошла ошибка при инциализации платежа</h3>';
            }
        }
		$this->setExtraParams($params);
        return $this->showTemplate($payment, "template");
	}
	/**
	 * @return array
	 */
	public static function getIndicativeFields()
	{
		return array('id', 'sum', 'clientid', 'orderid', 'key');
	}
    /** 
     * @param Request $request
     * @param $paySystemId
     * @return bool
     */
    static protected function isMyResponseExtended(Request $request, $paySystemId)
    {   
        $orderid = $request->get('orderid');
        $service_name = $request->get('service_name');
        if ($service_name == 'paykeeper_payment_system')
            return TRUE;
        $orderid_split = explode('|', $orderid);
        if (is_array($orderid_split))
            if (count($orderid_split) > 1)
                if ($orderid_split[1] == 'paykeeper_payment_system')
                    return TRUE;
        return FALSE;
    }
	/**
	 * @param Request $request
	 * @return mixed
	 */
	public function getPaymentIdFromRequest(Request $request)
	{
        return $this->getOrderIdFromRequest($request);
	}
	/**
	 * @param Payment $payment
	 * @param Request $request
	 * @return PaySystem\ServiceResult
	 */
	public function processRequest(Payment $payment, Request $request)
	{
        $payment_id = (int) $request->get('id');
        $clientid = $request->get('clientid');
        $sum = (float) $request->get('sum');
        $orderid = $request->get('orderid');
        $sign = $request->get('key');
        if($payment_id == 0)
        {
            echo "No payment specified!";
            die();
        }
		$busValues = $this->getParamsBusValue($payment);
        $secret = $busValues["PAYKEEPER_SECRET"];
        $hash = md5($payment_id.number_format($sum, 2, '.', '').$clientid.$orderid.$secret);
        if ($hash != $sign)
        {
            echo "Hash mismatch";
            die();
        }
        else
        {
            $result = new PaySystem\ServiceResult();
            $fields = array(
                "PS_STATUS" => "Y",
                "PS_STATUS_CODE" => "Success",
                "PS_STATUS_DESCRIPTION" => "Payment accepted",
                "PS_STATUS_MESSAGE" => "Payment id: $payment_id",
                "PS_SUM" => $sum,
                "PS_CURRENCY" => "",
            );
            $result->setOperationType(PaySystem\ServiceResult::MONEY_COMING);
            $result->setPsData($fields);
			//change order status according setting
            $status_id_after_payment = $busValues["PAYKEEPER_STATUS_AFTER_PAYMENT"];
            if ($status_id_after_payment != NULL) {
                if ((int) $orderid > 0) {
                    $order = \Bitrix\Sale\Order::load($orderid);
                    $r = $order->setField('STATUS_ID', $status_id_after_payment);
                    if (!$r->isSuccess())
                    {   
                        var_dump($r->getErrorMessages());
                    }   
                    else
                    {   
                        $r = $order->save();
                        if (!$r->isSuccess())
                        {   
                            var_dump($r->getErrorMessages());
                        }   
                        else
                            echo "OK ".md5($payment_id.$secret);
                    }   
                }   
            }   
            else
                echo "OK ".md5($payment_id.$secret);
        }
		return $result;
	}
	/**
	 * @return array
	 */
	public function getCurrencyList()
	{
		return array('RUB');
	}
    private function updateFiscalCart($name = "", $price = 0, $quantity = 0, $sum = 0, $tax = "none")
    {
        if ($this->getPaymentFormType() == "create") {
            $name = str_replace("\n ", "", $name);
            $name = str_replace("\r ", "", $name);
        }
        $this->fiscal_cart[] = array(
            "name" => $name,
            "price" => $price,
            "quantity" => $quantity,
            "sum" => $sum,
            "tax" => $tax
        );
    }
    private function setTaxes($sum, $tax_rate)
    {
        $taxes = array("tax" => "none", "tax_sum" => 0);
        switch($tax_rate) {
            case 10:
                $taxes["tax"] = "vat10";
                break;
            case 18: 
                $taxes["tax"] = "vat18";
                break;
            case 20: 
                $taxes["tax"] = "vat20";
                break;
        }   
        return $taxes;
    }
    private function getPaymentFormType()
    {
        if (strpos($this->getPaymentFormUrl(), "/order/inline") == True)
            return "order";
        else
            return "create";
    }
    private function setPaymentFormUrl($payment_form_url)
    {
        $this->payment_form_url = $payment_form_url;
    }
    private function getPaymentFormUrl()
    {
        return $this->payment_form_url;
    }
    private function getOrderIdFromRequest(Request $request)
    {
        if (strpos($request->get('orderid'), 'paykeeper_payment_system') !== false) {
            $order_data = explode("|", $request->get('orderid'));
            $orderid = $order_data[0];
        }
        else
            $orderid = $request->get('orderid');
        $order = Sale\Order::load($orderid);
        if (!is_null($order)) {
            return $order->getPaymentCollection()->current()->getField("ID");
        }
        else
            echo "Could not load order object by given orderid: $orderid";
    }
    private function correctPrecision()
    {
        //handle possible precision problem
        $fiscal_cart_sum = $this->getFiscalCartSum();
        $total_sum = $this->getOrderTotal();
        $diff_sum = $total_sum - $fiscal_cart_sum;
        //debug_info
        //echo "\ntotal: $total_sum - cart: $fiscal_cart_sum - diff: $diff_sum";
        if (abs($diff_sum) <= 0.01) {
            if ($this->getUseDelivery()) { //delivery is used
                $this->correctPriceOfCartItem($diff_sum, count($this->fiscal_cart)-1);
            }
            else {
                if ($this->single_item_index >= 0) { //we got single cart element
                    $this->correctPriceOfCartItem($diff_sum, $this->single_item_index);
                }
                else if ($this->more_then_one_item_index >= 0) { //we got cart element with more then one quantity
                    $this->splitCartItem($this->more_then_one_item_index);
                    //add diff_sum to the last element (just separated) of fiscal cart
                    $this->correctPriceOfCartItem($diff_sum, count($this->fiscal_cart)-1);
                }
                else { //we only got cart elements with less than one quantity
                    $modify_value = ($diff_sum > 0) ? $total_sum/$fiscal_cart_sum : $fiscal_cart_sum/$total_sum;
                    for ($pos=0; $pos<count($this->getFiscalCart()); $pos++) {
                        $this->fiscal_cart[$pos]["sum"] *= $modify_value;
                        $this->fiscal_cart[$pos]["price"] = $this->fiscal_cart[$pos]["sum"]/$this->fiscal_cart[$pos]["quantity"];
                    }
                }
            }
        }
    }
    private function setUseDelivery()
    {
        $this->use_delivery = True;
    }
    private function getUseDelivery()
    {
        return $this->use_delivery;
    }
    private function setOrderTotal($value)
    {
        $this->order_total = number_format($value, 2, ".", "");
    }
    private function getOrderTotal()
    {
        return $this->order_total;
    }
    private function getFiscalCartSum() {
        $fiscal_cart_sum = 0;
        foreach ($this->getFiscalCart() as $item) {
            $fiscal_cart_sum += floatval($item["sum"]);
        }
        return $fiscal_cart_sum;
    }
    private function correctPriceOfCartItem($corr_price_to_add, $item_position)
    {
        $item_sum = 0;
        $this->fiscal_cart[$item_position]["price"] += $corr_price_to_add;
        $item_sum = $this->fiscal_cart[$item_position]["price"]*$this->fiscal_cart[$item_position]["quantity"];
        $this->fiscal_cart[$item_position]["sum"] = $item_sum;
    }
    private function splitCartItem($cart_item_position)
    {
        $item_sum = 0;
        $item_price = 0;
        $item_quantity = 0;
        $item_price = $this->fiscal_cart[$cart_item_position]["price"];
        $item_quantity = $this->fiscal_cart[$cart_item_position]["quantity"]-1;
        $this->fiscal_cart[$cart_item_position]["quantity"] = $item_quantity; //decreese quantity by one
        $this->fiscal_cart[$cart_item_position]["sum"] = $item_price*$item_quantity; //new sum
        //add one cart item to the end of cart
        $this->updateFiscalCart(
            $this->fiscal_cart[$cart_item_position]["name"],
            $item_price, 1, $item_price,
            $this->fiscal_cart[$cart_item_position]["tax"]);
    }
    private function getFiscalCart()
    {
        return $this->fiscal_cart;
    }
}
